@@ -5,9 +5,9 @@ Split up into one class per command.
 """
 import click
 
-from .event import event
 from .list_disks import list_disks
-from .metadata import show_metadata
+from .metadata import metadata
+from .usercode import usercode
 
 
 @click.group("astdiskd")
@@ -15,9 +15,9 @@ def main() -> None:
     """Astoria Command Line Utility."""
 
 
-main.add_command(event)
 main.add_command(list_disks)
-main.add_command(show_metadata)
+main.add_command(metadata)
+main.add_command(usercode)
 
 if __name__ == "__main__":
     main()
