@@ -2,15 +2,20 @@
 
 import asyncio
 import logging
+from datetime import datetime
 from pathlib import Path
 from signal import SIGKILL, SIGTERM
 from tempfile import TemporaryDirectory
-from typing import Callable, Dict, Optional
+from typing import Callable, Dict, List, Optional, Set
 from zipfile import BadZipFile, ZipFile
 
 import click
+import toml
+from pydantic import ValidationError
 
 from astoria.common.broadcast_event import UsercodeLogBroadcastEvent
+from astoria.common.bundle import CodeBundle, IncompatibleKitVersionException
+from astoria.common.config import AstoriaConfig
 from astoria.common.manager import StateManager
 from astoria.common.manager_requests import (
     RequestResponse,
@@ -97,6 +102,7 @@ class ProcessManager(DiskHandlerMixin, StateManager[ProcessManagerMessage]):
                     disk_info,
                     self.update_status,
                     self._log_helper,
+                    self.config,
                 )
                 asyncio.ensure_future(self._lifecycle.run_process())
             else:
@@ -120,8 +126,8 @@ class ProcessManager(DiskHandlerMixin, StateManager[ProcessManagerMessage]):
                 LOGGER.warning("Disk removed, but no code lifecycle available")
 
     async def handle_kill_request(
-        self,
-        request: UsercodeKillManagerRequest,
+            self,
+            request: UsercodeKillManagerRequest,
     ) -> RequestResponse:
         """Handle a request to kill running usercode."""
         if self._lifecycle is None:
@@ -172,7 +178,7 @@ class ProcessManager(DiskHandlerMixin, StateManager[ProcessManagerMessage]):
         """
         if self._lifecycle is None:
             # When the status is updated in the lifecycle constructor, we
-            # are left with a situtation where there is no lifecycle object,
+            # are left with a situation where there is no lifecycle object,
             # but the code is starting. Thus we want to inform anyway.
             #
             # This section also updates the status when the lifecycle is cleaned up.
@@ -188,6 +194,10 @@ class ProcessManager(DiskHandlerMixin, StateManager[ProcessManagerMessage]):
             )
 
 
+class InvalidCodeBundleException(Exception):
+    """The code bundle was not valid."""
+
+
 class UsercodeLifecycle:
     """
     Manages the lifecycle of usercode.
@@ -196,16 +206,18 @@ class UsercodeLifecycle:
     """
 
     def __init__(
-        self,
-        uuid: DiskUUID,
-        disk_info: DiskInfo,
-        status_inform_callback: Callable[[CodeStatus], None],
-        log_helper: BroadcastHelper[UsercodeLogBroadcastEvent],
+            self,
+            uuid: DiskUUID,
+            disk_info: DiskInfo,
+            status_inform_callback: Callable[[CodeStatus], None],
+            log_helper: BroadcastHelper[UsercodeLogBroadcastEvent],
+            config: AstoriaConfig,
     ) -> None:
         self._uuid = uuid
         self._disk_info = disk_info
         self._status_inform_callback = status_inform_callback
         self._log_helper = log_helper
+        self._config = config
 
         self._process: Optional[asyncio.subprocess.Process] = None
         self._process_lock = asyncio.Lock()
@@ -239,7 +251,7 @@ class UsercodeLifecycle:
         self._dir = TemporaryDirectory(prefix="astprocd-")
         self._dir_path = Path(self._dir.name)
 
-    def _extract_and_validate_zip_file(self) -> bool:
+    def _extract_and_validate_zip_file(self) -> List[str]:
         """
         Extract and validate a robot.zip file.
 
@@ -247,42 +259,61 @@ class UsercodeLifecycle:
         robot.zip file. It will overwrite the current temporary dir
         of the usercode lifecycle, and should not be called multiple
         times.
+
+        :returns: messages to put in the log file
         """
-
-        def _handle_error(error: str) -> None:
-            """
-            Handle an error with the zip file.
-
-            Logs to the astprocd logger and also writes
-            a log file to the usercode disk.
-            """
-            with self._disk_info.mount_path.joinpath("log.txt").open("w") as fh:
-                LOGGER.warning(error)
-                fh.write("Unable to start code.\n")
-                fh.write(f"{error}.\n")
-
         zip_path = self._disk_info.mount_path / "robot.zip"
 
         if not (zip_path.exists() and zip_path.is_file()):
-            _handle_error("Unable to find robot.zip file")
-            return False
+            raise InvalidCodeBundleException("Unable to find robot.zip file")
 
         try:
             with ZipFile(zip_path, "r") as zf:
                 zf.extractall(self._dir_path)
         except BadZipFile:
-            _handle_error("The provided robot.zip is not a valid ZIP archive")
-            return False
+            raise InvalidCodeBundleException(
+                "The provided robot.zip is not a valid ZIP archive",
+            )
 
-        # Check that main.py is present
-        exe_path = self._dir_path / "main.py"
+        # Check that robot.py is present
+        exe_path = self._dir_path / "robot.py"
         if not exe_path.exists():
-            _handle_error("The provided robot.zip did not contain a main.py")
-            return False
+            raise InvalidCodeBundleException(
+                "The provided robot.zip did not contain a robot.py",
+            )
 
-        # Check for metadata etc here in future
+        LEGACY_FILES: Set[str] = {"info.yaml", "info.yml", "wifi.yaml", "wifi.yml"}
 
-        return True
+        for legacy_file in LEGACY_FILES:
+            legacy_info_path = self._dir_path / legacy_file
+            if legacy_info_path.exists():
+                raise InvalidCodeBundleException(
+                    "This is an old robot code package and \
+                    will not work with this version of the kit.",
+                )
+
+        # Check for code bundle info
+        bundle_meta_path = self._dir_path / "bundle.toml"
+        if not bundle_meta_path.exists():
+            raise InvalidCodeBundleException(
+                "The provided robot.zip did not contain a bundle.toml",
+            )
+        else:
+            # Validate code bundle info
+            try:
+                bundle_contents = toml.loads(bundle_meta_path.read_text())
+            except toml.TomlDecodeError as e:
+                raise InvalidCodeBundleException(f"The code bundle was invalid.\n{e}")
+
+            try:
+                bundle = CodeBundle(**bundle_contents)
+            except ValidationError as e:
+                raise InvalidCodeBundleException(f"The code bundle was invalid.\n{e}")
+
+            try:
+                return bundle.check_kit_version_is_compatible(self._config.kit)
+            except IncompatibleKitVersionException as e:
+                raise InvalidCodeBundleException(f"Invalid code bundle: {e}")
 
     async def run_process(self) -> None:
         """
@@ -291,14 +322,13 @@ class UsercodeLifecycle:
         This function will not return until the code has exited.
         """
         if self._process is None:
-
-            if self._extract_and_validate_zip_file():
-
+            try:
+                log_message = self._extract_and_validate_zip_file()
                 async with self._process_lock:
                     self._process = await asyncio.create_subprocess_exec(
                         "python3",
                         "-u",
-                        "main.py",
+                        "robot.py",
                         stdin=asyncio.subprocess.DEVNULL,
                         stdout=asyncio.subprocess.PIPE,
                         stderr=asyncio.subprocess.STDOUT,
@@ -307,7 +337,12 @@ class UsercodeLifecycle:
                     )
                     if self._process is not None:
                         if self._process.stdout is not None:
-                            asyncio.ensure_future(self.logger(self._process.stdout))
+                            asyncio.ensure_future(
+                                self.logger(
+                                    self._process.stdout,
+                                    initial_messages=log_message,
+                                ),
+                            )
                         else:
                             LOGGER.warning("Unable to start logger task.")
                         self.status = CodeStatus.RUNNING
@@ -336,10 +371,16 @@ class UsercodeLifecycle:
                         self._dir.cleanup()  # Reset directory
                         self._setup_temp_dir()
                     else:
-                        LOGGER.warning("robot.zip was invalid. Unable to start code.")
+                        LOGGER.warning("Tried to start process, but failed.")
                         self.status = CodeStatus.CRASHED  # Close enough to indicate error
-            else:
-                LOGGER.warning("Tried to start process, but failed.")
+            except InvalidCodeBundleException as e:
+                LOGGER.warning("robot.zip was invalid. Unable to start code.")
+                # Write error message to the log file.
+                with self._disk_info.mount_path.joinpath("log.txt").open("w") as fh:
+                    LOGGER.warning(str(e))
+                    fh.write("Unable to start code.\n")
+                    fh.write(f"{e}.\n")
+
                 self.status = CodeStatus.CRASHED  # Close enough to indicate error
         else:
             LOGGER.warning("Tried to start process, but one is already running.")
@@ -364,11 +405,19 @@ class UsercodeLifecycle:
         else:
             LOGGER.debug("Tried to kill process, but no process is running.")
 
-    async def logger(self, proc_output: asyncio.StreamReader) -> None:
+    async def logger(
+        self,
+        proc_output: asyncio.StreamReader,
+        *,
+        initial_messages: List[str] = [],
+    ) -> None:
         """
         Logger task.
 
-        Logs the output of the process to log.txt
+        Logs the output of the process to a log file and MQTT
+
+        :param proc_output: stream of data from the usercode process
+        :param initial_message: optional message to add at the start of the log
         """
         log_path = self._disk_info.mount_path / "log.txt"
 
@@ -388,14 +437,27 @@ class UsercodeLifecycle:
 
         with log_path.open("w") as fh:
             log_line = 0
-            log("=== LOG STARTED ===\n", log_line)
+
+            start_time = datetime.now()
+
+            # Print any initial messages
+            for message in initial_messages:
+                time_passed = datetime.now() - start_time
+                log(f"[{time_passed}] {message}\n", log_line)
+                log_line += 1
+
+            time_passed = datetime.now() - start_time
+            log(f"[{time_passed}] === LOG STARTED ===\n", log_line)
             log_line += 1
+
             data = await proc_output.readline()
             while data != b"":
-                log(data.decode('utf-8'), log_line)
+                data_str = data.decode('utf-8')
+                time_passed = datetime.now() - start_time
+                log(f"[{time_passed}] {data_str}", log_line)
                 data = await proc_output.readline()
                 log_line += 1
-            log("=== LOG FINISHED ===\n", log_line)
+            log(f"[{time_passed}] === LOG FINISHED ===\n", log_line)
 
 
 if __name__ == "__main__":
